import typing as t

from jinja2 import nodes
from jinja2.compiler import CodeGenerator
from jinja2.compiler import CompilerExit
from jinja2.compiler import Frame


class AsyncCodeGenerator(CodeGenerator):
    def visit_Block(self, node: nodes.Block, frame: Frame) -> None:
        level = 0
        if frame.toplevel:
            if self.has_known_extends:
                return
            if self.extends_so_far > 0:
                self.writeline("if parent_template is None:")
                self.indent()
                level += 1
        if node.scoped:
            context = self.derive_context(frame)
        else:
            context = self.get_context_ref()
        if node.required:
            self.writeline(f"if len(context.blocks[{node.name!r}]) <= 1:", node)
            self.indent()
            self.writeline(
                f'raise TemplateRuntimeError("Required block {node.name!r} not found")',
                node,
            )
            self.outdent()
        if not self.environment.is_async and frame.buffer is None:
            self.writeline(
                f"yield from context.blocks[{node.name!r}][0]({context})", node
            )
        else:
            self.writeline(
                f"{self.choose_async()}for event in"
                f" context.blocks[{node.name!r}][0]({context}):",
                node,
            )
            self.indent()
            self.simple_write("event", frame)
            self.outdent()
        self.outdent(level)

    def visit_Extends(self, node: nodes.Extends, frame: Frame) -> None:
        if not frame.toplevel:
            self.fail("cannot use extend from a non top-level scope", node.lineno)
        if self.extends_so_far > 0:
            if not self.has_known_extends:
                self.writeline("if parent_template is not None:")
                self.indent()
            self.writeline('raise TemplateRuntimeError("extended multiple times")')
            if self.has_known_extends:
                raise CompilerExit()
            else:
                self.outdent()
        self.writeline(
            f"parent_template = "
            f"{self.choose_async('await ')}environment.get_template(",
            node,
        )
        self.visit(node.template, frame)
        self.write(f", {self.name!r})")
        self.writeline("for name, parent_block in parent_template.blocks.items():")
        self.indent()
        self.writeline("context.blocks.setdefault(name, []).append(parent_block)")
        self.outdent()
        if frame.rootlevel:
            self.has_known_extends = True
        self.extends_so_far += 1

    def visit_Include(self, node: nodes.Include, frame: Frame) -> None:
        if node.ignore_missing:
            self.writeline("try:")
            self.indent()
        func_name = "get_or_select_template"
        if isinstance(node.template, nodes.Const):
            if isinstance(node.template.value, str):
                func_name = "get_template"
            elif isinstance(node.template.value, (tuple, list)):
                func_name = "select_template"
        elif isinstance(node.template, (nodes.Tuple, nodes.List)):
            func_name = "select_template"
        self.writeline(
            f"template = {self.choose_async('await ')}environment.{func_name}(", node
        )
        self.visit(node.template, frame)
        self.write(f", {self.name!r})")
        if node.ignore_missing:
            self.outdent()
            self.writeline("except TemplateNotFound:")
            self.indent()
            self.writeline("pass")
            self.outdent()
            self.writeline("else:")
            self.indent()
        skip_event_yield = False
        if node.with_context:
            self.writeline(
                f"{self.choose_async()}for event in template.root_render_func("
                "template.new_context(context.get_all(), True,"
                f" {self.dump_local_context(frame)})):"
            )
        elif self.environment.is_async:
            self.writeline(
                "for event in (await template._get_default_module_async())"
                "._body_stream:"
            )
        else:
            self.writeline("yield from template._get_default_module()._body_stream")
            skip_event_yield = True
        if not skip_event_yield:
            self.indent()
            self.simple_write("event", frame)
            self.outdent()
        if node.ignore_missing:
            self.outdent()

    def _import_common(
        self, node: t.Union[nodes.Import, nodes.FromImport], frame: Frame
    ) -> None:
        self.write(f"{self.choose_async('await ')}environment.get_template(")
        self.visit(node.template, frame)
        self.write(f", {self.name!r}).")
        if node.with_context:
            f_name = f"make_module{self.choose_async('_async')}"
            self.write(
                f"{f_name}(context.get_all(), True, {self.dump_local_context(frame)})"
            )
        else:
            self.write(f"_get_default_module{self.choose_async('_async')}(context)")



import importlib.util
import typing as t
import zipimport
from collections import abc
from contextlib import suppress
from importlib import import_module

from aiopath import AsyncPath
from jinja2.environment import Template
from jinja2.exceptions import TemplateNotFound
from jinja2.utils import internalcode
from .environment import AsyncEnvironment


class PackageSpecNotFound(TemplateNotFound):
    """Raised if a package spec not found."""


class LoaderNotFound(TemplateNotFound):
    """Raised if a loader is not found."""


class AsyncBaseLoader:
    has_source_access = True

    def __init__(self, searchpath: AsyncPath | t.Sequence[AsyncPath]) -> None:
        self.searchpath = searchpath
        if not isinstance(searchpath, abc.Iterable):
            self.searchpath = [searchpath]

    async def get_source(
        self, environment: "AsyncEnvironment", template: AsyncPath
    ) -> t.Tuple[
        str, t.Optional[AsyncPath], t.Optional[t.Callable[[], bool]]
    ] | t.NoReturn:
        if not self.has_source_access:
            raise RuntimeError(
                f"{type(self).__name__} cannot provide access to the source"
            )
        raise TemplateNotFound(template.name)

    async def list_templates(self) -> list[str] | t.NoReturn:
        raise TypeError("this loader cannot iterate over all templates")

    @internalcode
    async def load(
        self,
        environment: "AsyncEnvironment",
        name: str,
        env_globals: t.Optional[t.MutableMapping[str, t.Any]] = None,
    ) -> "Template":
        if env_globals is None:
            env_globals = {}
        code: t.Any = None
        bucket: t.Any = None
        source, path, uptodate = await self.get_source(environment, AsyncPath(name))
        bcc = environment.bytecode_cache
        if bcc:
            bucket = await bcc.get_bucket(environment, name, path, source)
            code = bucket.code
        if not code:
            code = environment.compile(source, name, path.name)  # type: ignore
        if bcc and not bucket.code:
            bucket.code = code
            await bcc.set_bucket(bucket)
        return environment.template_class.from_code(
            environment, code, env_globals, uptodate
        )


class FileSystemLoader(AsyncBaseLoader):
    def __init__(
        self,
        searchpath: AsyncPath | t.Sequence[AsyncPath],
        encoding: str = "utf-8",
        followlinks: bool = False,
    ) -> None:
        super().__init__(searchpath)
        self.encoding = encoding
        self.followlinks = followlinks

    async def get_source(
        self, environment: "AsyncEnvironment", template: AsyncPath
    ) -> tuple[str, AsyncPath, t.Callable | bool]:
        # split_template_path(template)
        for searchpath in self.searchpath:  # type: ignore
            path = searchpath / template
            if await path.is_file():
                break
        else:
            raise TemplateNotFound(template.name)
        try:
            resp = await path.read_text()
        except FileNotFoundError:
            raise TemplateNotFound(path.name)
        mtime = (await path.stat()).st_mtime

        async def uptodate() -> bool:
            try:
                return (await path.stat()).st_mtime == mtime
            except OSError:
                return False

        return resp, path, uptodate

    async def list_templates(self) -> list[str]:
        found = set()
        for searchpath in self.searchpath:  # type: ignore
            found.update([path.name async for path in searchpath.rglob("*.html")])
        return sorted(found)


class PackageLoader(AsyncBaseLoader):
    def __init__(
        self,
        package_name: str,
        searchpath: AsyncPath | t.Sequence[AsyncPath],
        package_path: AsyncPath = AsyncPath("templates"),
        encoding: str = "utf-8",
    ) -> None:
        super().__init__(searchpath)
        self.package_path = package_path
        self.package_name = package_name
        self.encoding = encoding
        # Make sure the package exists. This also makes namespace
        # packages work, otherwise get_loader returns None.
        import_module(package_name)
        spec = importlib.util.find_spec(package_name)
        if not spec:
            raise PackageSpecNotFound("An import spec was not found for the package")
        loader = spec.loader
        if not loader:
            raise LoaderNotFound("A loader was not found for the package")
        self._loader = loader
        self._archive = None
        template_root = None
        if isinstance(loader, zipimport.zipimporter):
            self._archive = loader.archive
            pkgdir = next(iter(spec.submodule_search_locations))  # type: ignore
            template_root = AsyncPath(pkgdir) / package_path
        else:
            roots = []
            # One element for regular packages, multiple for namespace
            # packages, or None for single module file.
            if spec.submodule_search_locations:
                roots.extend([AsyncPath(s) for s in spec.submodule_search_locations])
            # A single module file, use the parent directory instead.
            elif spec.origin is not None:
                roots.append(AsyncPath(spec.origin))
            for root in roots:
                path = root / package_path
                if path.is_dir():
                    template_root = root
                    break

        if not template_root:
            raise ValueError(
                f"The {package_name!r} package was not installed in a"
                " way that PackageLoader understands"
            )

        self._template_root = template_root

    async def get_source(
        self, environment: "AsyncEnvironment", template: AsyncPath
    ) -> t.Tuple[str, AsyncPath, t.Optional[t.Coroutine[t.Any, t.Any, bool]]]:
        path = self._template_root / template
        # up_to_date: t.Optional[t.Callable[[], bool]] | None
        if self._archive:
            if not path.is_file():
                raise TemplateNotFound(path.name)
            source = await path.read_bytes()
            mtime = (await path.stat()).st_mtime

            async def up_to_date() -> bool:
                return await path.is_file() and (await path.stat()).st_mtime == mtime

        else:
            # Package is a zip file.
            try:
                source = self._loader.get_data(str(path))  # type: ignore
            except OSError as e:
                raise TemplateNotFound(path.name) from e
            up_to_date = None  # type: ignore
        return source.decode(self.encoding), path, up_to_date  # type: ignore

    async def list_templates(self) -> list[AsyncPath]:
        results: list[AsyncPath] = []

        if self._archive is None:
            # Package is a directory.
            paths = self._template_root.rglob("*.html")
            results.extend([p async for p in paths])
        else:
            if not hasattr(self._loader, "_files"):
                raise TypeError(
                    "This zip import does not have the required"
                    " metadata to list templates"
                )
            # Package is a zip file.
            prefix = self._template_root.name
            for name in self._loader._files.keys():  # type: ignore
                # Find names under the templates directory that aren't directories.
                if name.startswith(prefix) and (await AsyncPath(name).is_file()):
                    results.append(name)
        results.sort()
        return results


class DictLoader(AsyncBaseLoader):
    def __init__(
        self,
        mapping: t.Mapping[str, str],
        searchpath: AsyncPath | t.Sequence[AsyncPath],
    ) -> None:
        super().__init__(searchpath)
        self.mapping = mapping

    async def get_source(
        self, environment: "AsyncEnvironment", template: str
    ) -> t.Tuple[str, None, t.Callable[[], bool]]:
        if template in self.mapping:
            source = self.mapping[template]
            return source, None, lambda: source == self.mapping.get(template)
        raise TemplateNotFound(template)

    async def list_templates(self) -> list[str]:
        return sorted(self.mapping)


class FunctionLoader(AsyncBaseLoader):
    def __init__(
        self,
        load_func: t.Callable,
        searchpath: AsyncPath | t.Sequence[AsyncPath],
    ) -> None:
        super().__init__(searchpath)
        self.load_func = load_func

    async def get_source(
        self, environment: "AsyncEnvironment", template: str | AsyncPath
    ) -> t.Tuple[str, AsyncPath | None, t.Callable | bool | None]:
        path = AsyncPath(template)
        source = self.load_func(path)
        if source is None:
            raise TemplateNotFound(path.name)
        if isinstance(source, str):
            return source, path, True
        return source


class ChoiceLoader(AsyncBaseLoader):
    loaders: list[AsyncBaseLoader] = []

    def __init__(
        self,
        loaders: list[AsyncBaseLoader],
        searchpath: AsyncPath | t.Sequence[AsyncPath],
    ) -> None:
        super().__init__(searchpath)
        self.loaders = loaders

    async def get_source(
        self, environment: "AsyncEnvironment", template: AsyncPath
    ) -> t.Tuple[str, t.Optional[AsyncPath], t.Optional[t.Callable | bool]]:
        for loader in self.loaders:
            with suppress(TemplateNotFound):
                return await loader.get_source(environment, template)
        raise TemplateNotFound(template.name)

    async def list_templates(self) -> list[str]:
        found = set()
        for loader in self.loaders:
            found.update(await loader.list_templates())
        return sorted(found)



from .bccache import AsyncRedisBytecodeCache
from .environment import AsyncEnvironment
from .loaders import AsyncBaseLoader
from .loaders import ChoiceLoader
from .loaders import DictLoader
from .loaders import FileSystemLoader
from .loaders import FunctionLoader
from .loaders import PackageLoader

__all__ = [
    "AsyncEnvironment",
    "AsyncBaseLoader",
    "FunctionLoader",
    "FileSystemLoader",
    "PackageLoader",
    "DictLoader",
    "ChoiceLoader",
    "AsyncRedisBytecodeCache",
]



import typing as t
from weakref import ref

from jinja2 import Environment
from jinja2.environment import Template
from jinja2.utils import internalcode
from .compiler import AsyncCodeGenerator


class AsyncEnvironment(Environment):
    code_generator_class = AsyncCodeGenerator
    loader: t.Optional[t.Any] = None
    bytecode_cache: t.Optional[t.Any] = None

    @internalcode
    async def get_template(
        self,
        name: str,
        parent: t.Optional[str] = None,
        env_globals: t.Optional[t.MutableMapping[str, t.Any]] = None,
    ) -> "Template":
        if isinstance(name, Template):
            return name
        if parent is not None:
            name = self.join_path(name, parent)
        return await self._load_template(name, env_globals)

    @internalcode
    async def _load_template(
        self, name: str, env_globals: t.Optional[t.MutableMapping[str, t.Any]]
    ) -> "Template":
        if self.loader is None:
            raise TypeError("no loader for this environment specified")
        cache_key = (ref(self.loader), name)
        if self.cache is not None:
            template = self.cache.get(cache_key)
            if template and (not self.auto_reload or template.is_up_to_date):
                if env_globals:
                    template.globals.update(env_globals)
                return template
        template = await self.loader.load(self, name, self.make_globals(env_globals))
        if self.cache is not None:
            self.cache[cache_key] = template
        return template



import typing as t

from aiopath import AsyncPath
from jinja2 import BytecodeCache
from jinja2.bccache import Bucket
from redis.asyncio import Redis as AsyncRedis
from .environment import AsyncEnvironment


class AsyncRedisBytecodeCache(AsyncRedis, BytecodeCache):
    def __init__(self, prefix: t.Optional[str] = None, **configs) -> None:
        super().__init__(**configs)
        self.prefix = prefix

    def get_bucket_name(self, key: str) -> str:
        return ":".join([self.prefix, key]) if self.prefix else key

    async def load_bytecode(self, key_bucket: Bucket) -> bytes | None:
        code = await self.get(self.get_bucket_name(key_bucket.key))
        if code:
            return key_bucket.bytecode_from_string(code)

    async def dump_bytecode(self, key_bucket: Bucket) -> None:
        await self.set(
            self.get_bucket_name(key_bucket.key), key_bucket.bytecode_to_string()
        )

    async def get_bucket(
        self,
        environment: "AsyncEnvironment",
        name: str,
        path: AsyncPath,
        source: str,
    ) -> Bucket:
        key = self.get_cache_key(name, path.name)
        checksum = self.get_source_checksum(source)
        bucket = Bucket(environment, key, checksum)
        await self.load_bytecode(bucket)
        return bucket

    async def set_bucket(self, bucket: Bucket) -> None:
        await self.dump_bytecode(bucket)
